#
#   PROGRAM:     Power Console - Firebird QA toolset
#   MODULE:      fbqacmd.py
#   DESCRIPTION: PowerConsole commands for Firebird QA
#
#  The contents of this file are subject to the Initial
#  Developer's Public License Version 1.0 (the "License");
#  you may not use this file except in compliance with the
#  License. You may obtain a copy of the License at
#  http://www.firebirdsql.org/index.php?op=doc&id=idpl
#
#  Software distributed under the License is distributed AS IS,
#  WITHOUT WARRANTY OF ANY KIND, either express or implied.
#  See the License for the specific language governing rights
#  and limitations under the License.
#
#  The Original Code was created by Pavel Cisar
#
#  Copyright (c) 2006-2009 Pavel Cisar <pcisar@users.sourceforge.net>
#  and all contributors signed below.
#
#  All Rights Reserved.
#  Contributor(s): ______________________________________.

from pwc.base import *
from pyparsing import *
from xml.dom import minidom
import os
import sys
from configobj import ConfigObj
import platform
import tempfile
from Cheetah.Template import Template
import glob
import kinterbasdb as kdb
#if not kdb.initialized:
    #kdb.init(type_conv=300)
import kinterbasdb.services as fbservice
import gc
import release

try:
    import qm
except ImportError:
    qm_present = False
else:
    qm_present = True
    import qm
    import qm.xmlutil
    from qm.test import base
    from qm.test.database import TestDescriptor
    from qm.test.result import Result
    from qm.test.classes.xml_database import XMLDatabase
    from qm.extension import get_class_arguments
    from qm.test.cmdline import QMTest
    qm.common.program_name = "QMTest"
    qm.diagnostic.load_messages("test")
    qm.rc.Load("test")
    if sys.platform != 'win32':
        qm.sigmask.save_mask()

UNKNOWN = 'Unknown'
QM_ERR  = "QMTest not installed.\nYou can download QMTest from www.qmtest.com.\n"

class Report(object):
    """Encapsulates test results from all targets and runs.
    Attributes:
    id              Test/resource etc. ID.
    kind            Kind of result (test, resource etc.)
    suite           List of all test IDs.
    targets         List of TargetResult instances.
    header          List of report HEADER data (target,(seqno,seqno,..))
    data            Dictionary of report DATA ((result,result,..),..) keyed by test ID.
    """

    css = """body {
    background: fff;
    color: 000;
    margin: 10px;
    }
    body, th, td {
    font: normal 13px verdana,arial,'Bitstream Vera Sans',helvetica,sans-serif;
    }
    h1, h2, h3, h4 {
    font-family: arial,verdana,'Bitstream Vera Sans',helvetica,sans-serif;
    font-weight: bold;
    letter-spacing: -0.018em;
    }
    h1 { font-size: 21px; margin: .15em 1em .15em 0 }
    h2 { font-size: 16px; margin: 2em 0 .5em; }
    h3 { font-size: 14px; margin: 1.5em 0 .5em; }
    hr { border: none;  border-top: 1px solid ccb; margin: 2em 0; }
    address { font-style: normal }
    img { border: none }
   
    .underline { text-decoration: underline; }
    ol.loweralpha { list-style-type: lower-alpha }
    ol.upperalpha { list-style-type: upper-alpha }
    ol.lowerroman { list-style-type: lower-roman }
    ol.upperroman { list-style-type: upper-roman }
    ol.arabic     { list-style-type: decimal }
   
    /* Link styles */
    :link, :visited {
    text-decoration: none;
    color: black;
    border-bottom: 1px dotted blue;
    }
   :link:hover, :visited:hover {
    background-color: eee;
    color: black;
    border-bottom: 1px solid blue;
    }
    h1 :link, h1 :visited ,h2 :link, h2 :visited, h3 :link, h3 :visited,
    h4 :link, h4 :visited, h5 :link, h5 :visited, h6 :link, h6 :visited {
    color: inherit;
    }
   
    .ext-link { background: url(../extlink.gif) no-repeat 0 58%; padding-left: 16px }
    * html .ext-link { background-position: 0 .35em } /* IE hack, see 937 */
   
    .menuitem { background-color: rgb(0, 153, 0); font-weight: bold; }
   
    .menuitem :link, .menuitem :visited {
       text-decoration: none;
       color: white;
       border-bottom: none;
       }
   
    .menuitem :link:hover, .menuitem :visited:hover {
       text-decoration: none;
       color: black;
       border-bottom: none;
    }
   
    .menuspacer { font-weight: bold; margin: 2em 0 .5em; }
   
    .header { 
       background-color: rgb(0, 153, 0);
       color: rgb(255, 255, 255);
       white-space: nowrap; 
       text-align: center;
    }
   
    .topiclist { background-color: eee; }
    td.listifo   { background-color: eee; }
   
    div { margin: 1.0em 0 1.0em; }
   
    div.pagination { font-size: 14px; margin: 1.0em 0 1.0em; font-weight: bold; }
    div.messageheader { background-color: eee; }
    div.messagebody { background-color: white; }
    div.threadmap { font-size: 14px; margin: 1.0em 0 1.0em; }
   
    table.summary {text-align: left; }
    td.sumPASS      { background-color: #a0f0a0;}
    td.sumFAIL      { background-color: #ff9090;}
    td.sumUNTESTED  { background-color: d0d0d0;}
    td.sumERROR     { background-color: #ffffe0;}
    td.sumline         { background-color: eee; }
   
    table.detail   { }

    td.annotation { text-align: left; color: black; background-color: eee; }
    td.value      { text-align: left; color: black; background-color: eee; }
   
    td.header    { text-align: center; color: black; font-weight: bold; background-color: #e0e0e0; }
    td.line      { text-align: left; color: black; background-color: eee; }
    td.PASS      { text-align: center; background-color: #a0f0a0;}
    td.FAIL      { text-align: center; background-color: #ff9090;}
    td.UNTESTED  { text-align: center; background-color: d0d0d0;}
    td.ERROR     { text-align: center; background-color: #ffffe0;}
    td.NA        { text-align: center; background-color: eee;}
    """
    
    tmplDetail = """$pageHeader.respond
    <H1 class="header">Details for $testName</H1>
    #for $name,$annotations in $outcomes
    <H2>$name</H2>
    <table class="detail" width="100%" cellpadding="2" cellspacing="2">
    #for $annotation in $annotationOrder
    #if $annotations.has_key($annotation)
    <tr>
    <td class="annotation" width="200">$annotation</td>
    <td class="value" width="%80">$annotations[$annotation]</td>
    </tr>
    #end if
    #end for
    #for $k in $annotations.keys()
    #if $k not in $annotationOrder
    <tr>
    <td class="annotation" width="200">$k</td>
    <td class="value" width="%80">$annotations[$k]</td>
    </tr>
    #end if
    #end for
    </table>
    #end for
    $pageFooter.respond
    """
    
    tmplMain = """$pageHeader.respond
    <H1 class="header">Firebird $version - QA Report</H1>
    <H2>Summary</H2>
    <table width="200">
    <tr><td class="sumPASS">Passed</td><td class="sumPASS" align="right">$passCount</td></tr>
    <tr><td class="sumFAIL">Failed</td><td class="sumFAIL" align="right">$failCount</td></tr>
    <tr><td class="sumUNTESTED">Untested</td><td class="sumUNTESTED" align="right">$untestedCount</td></tr>
    <tr><td class="sumERROR">Errors</td><td class="sumERROR" align="right">$errorCount</td></tr>
    <tr><td class="sumline">Total</td><td class="sumline" align="right">$totalCount</td></tr>
    </table>
    #if len($_report) > 0
    <H2>Tests with issues</H2>
    #set $testno = 1
    <table  width="100%" cellpadding="2" cellspacing="2">
    $tableHeader.respond()
    #for $test in $_report
    <tr>
    #set $filename = 'detail_%03i.html' % $testno
    <td class="line">$testno</td><td class="line"><a href="$filename">$test</a></td>
    #for $target in $data[$test]
      #for $result in $target
        #if $result != None:
    <td class="$result.GetOutcome()">$result.GetOutcome()</td>
        #else:
    <td class="NA">N/A</td>
        #end if
      #end for
    #end for
    </tr>
    #set $testno = $testno+1
    #if ($testno % 20) == 0
    $tableHeader.respond()
    #end if
    #end for
    </table>
    #end if
    $pageFooter.respond
    """
    
    tmplTableHeader = """#set $targetWidth = 65 / len($header)
    <tr>
    <td class="header" rowspan="2" width="%5">No.</td><td class="header" rowspan="2" width="%30">Test name</td>
    #for $target,$span in $header
    <td class="header" width="%$targetWidth" colspan="$len($span)">$target</td>
    #end for
    </tr>
    <tr>
    #for $target,$span in $header
      #for $run in $span
    <td class="header">$run</td>
      #end for
    #end for
    </tr>
    """
    
    tmplHeader = """#errorCatcher Echo
    <!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
    <html>
    <head>
      <meta content="text/html; charset=ISO-8859-1" http-equiv="content-type">
      <title>Firebird QA Report</title>
      <style type="text/css">
    @import url(css.css);
      </style>
    </head>
    <body>
    <a name="top"></a>
    """
    
    tmplFooter = """</body>
    </html>
    """
    
    class detailContext(object):
        """Helper class.
        """
        testName = ""
        annotationOrder = ['outcome','qmtest.cause'
            'qmtest.exception',
            'qmtest.resource',
            'qmtest.traceback',
            'FirebirdTest.ISQL_stripped_diff',
            'FirebirdTest.ISQL_stdout_actual',
            'FirebirdTest.ISQL_stdout_expected',
            'FirebirdTest.ISQL_stdout_actual_stripped',
            'FirebirdTest.ISQL_stdout_expected_stripped',
            'FirebirdTest.ISQL_stderr_stripped_diff',
            'FirebirdTest.ISQL_stderr_actual',
            'FirebirdTest.ISQL_stderr_expected',
            'FirebirdTest.ISQL_stderr_actual_stripped',
            'FirebirdTest.ISQL_stderr_expected_stripped',
            'FirebirdTest.python_stripped_diff',
            'FirebirdTest.python_stdout_actual',
            'FirebirdTest.python_stdout_expected',
            'FirebirdTest.python_stdout_actual_stripped',
            'FirebirdTest.python_stdout_expected_stripped',
            'FirebirdTest.ISQL_STDERR',
            ]
        outcomes = []
        
    def __init__(self,targets,version):
        self._report = []
        self.targets = targets
        # Build list of all tests and header data
        suiteDict = {}
        self.header = []
        for target in targets:
            self.header.append((target.name,tuple([run.seqno for run in target.runs])))
            for run in target.runs:
                suiteDict.update(run.results)
        self.suite = suiteDict.keys()
        self.suite.sort()
        # Build report data
        self.data = {}
        for test in self.suite:
            self.data[test] = tuple([tuple([run.results.get(test,None) for run in target.runs]) for target in targets])
        # summaries
        self.totalCount = len(self.suite)
        self.failCount = self._countOutcome('FAIL')
        self.untestedCount = self._countOutcome('UNTESTED')
        self.errorCount = self._countOutcome('ERROR')
        self.passCount = self.totalCount - self.untestedCount - self.errorCount - self.failCount
        self.version = version

        # HTML templates
        self.__HTMLNamespace = {}
        self.detailContext = self.detailContext()
        self.__tmplHeader = Template(self.tmplHeader,searchList=[self.__HTMLNamespace,self])
        self.__tmplFooter = Template(self.tmplFooter,searchList=[self.__HTMLNamespace,self])
        self.__tmplTableHeader = Template(self.tmplTableHeader,searchList=[self.__HTMLNamespace,self])
        self.__tmplMain = Template(self.tmplMain,searchList=[self.__HTMLNamespace,self])
        self.__tmplDetail = Template(self.tmplDetail,searchList=[self.__HTMLNamespace,self, self.detailContext])
        self.__HTMLNamespace["pageHeader"] = self.__tmplHeader
        self.__HTMLNamespace["pageFooter"] = self.__tmplFooter
        self.__HTMLNamespace["tableHeader"] = self.__tmplTableHeader

    def _TestDataAsList(self,testData):
        l = []
        for run in testData:
            for result in run:
                l.append(result)
        return l
    def _countOutcome(self,outcome):
        count = 0
        def checkOutcome(targets):
            for run in targets:
                for result in run:
                    if (result != None) and (result.GetOutcome() == outcome):
                        return True
            return False
        for test,targets in self.data.items():
            if checkOutcome(targets):
                count += 1
        return count
    def _outcomeValue(self,result):
        try:
            return ['PASS','UNTESTED','ERROR','FAIL'].index(result.GetOutcome())
        except AttributeError, ValueError:
            return -1
    def _getWorstOutcome(self,testData):
        seen = -1
        for result in self._TestDataAsList(testData):
            value = self._outcomeValue(result)
            if seen < value:
                seen = value
        return seen
    def _compareTestResults(self,testA,testB):
        valueA = self._getWorstOutcome(self.data[testA])
        valueB = self._getWorstOutcome(self.data[testB])
        value = valueB - valueA
        if value == 0:
            return cmp(testA,testB)
        else:
            return value
    def _resultsAreEqual(self,rA,rB):
        if rA.GetOutcome() != rB.GetOutcome():
            return False
        for k in [k for k in rA.keys() if k not in ['qmtest.end_time',
            'qmtest.start_time','qmtest.target','FirebirdTest.ISQL_stdout_actual',
            'FirebirdTest.ISQL_stderr_actual','FirebirdTest.python_stdout_actual']]:
            if rA[k] != rB.get(k,None):
                return False
        for k in [k for k in rB.keys() if k not in ['qmtest.end_time',
            'qmtest.start_time','qmtest.target','FirebirdTest.ISQL_stdout_actual',
            'FirebirdTest.ISQL_stderr_actual','FirebirdTest.python_stdout_actual']]:
            if rB[k] != rA.get(k,None):
                return False
        return True
    def _groupResults(self,testData):
        """Returns list of results grouped by the same key annotations.
        List structure:
        [[result,..],..]
        """
        resultList = []
        for targetRuns in testData:
            for runResult in targetRuns:
                if runResult != None:
                    found = False
                    # loop over resultList groups
                    for resultGroup in resultList:
                        if self._resultsAreEqual(runResult,resultGroup[0]):
                            resultGroup.append(runResult)
                            found = True
                            break
                    if not found:
                        # Not found in any group, create new one
                        resultList.append([runResult])
        return resultList

    def generateHTMLReport(self,path,full=False):
        if full:
            self._report = self.suite.copy()
        else:
            self._report = [r for r in self.suite if self._getWorstOutcome(self.data[r]) != 0]
        self._report.sort(lambda testA, testB: self._compareTestResults(testA,testB))

        f = file(os.path.join(path,'css.css'),'w')
        f.write(self.css)
        f.close()
        f = file(os.path.join(path,'index.html'),'w')
        f.write(self.__tmplMain.respond())
        f.close()
        # detail pages
        i = 1
        for test in self._report:
            self.detailContext.testName = test
            self.detailContext.outcomes = []
            # build list of unique annotation sets
            for resultGroup in self._groupResults(self.data[test]):
                # Construct list of targets and runs
                names = []
                for result in resultGroup:
                    names.append(' '.join((result.owner.target,result.owner.seqno)))
                # Create list of annotations
                annotations = {}
                c = resultGroup[0].GetCause()
                if c != "":
                    annotations[resultGroup[0].CAUSE] = c
                annotations["outcome"] = resultGroup[0].GetOutcome()
                for k,v in resultGroup[0].items():
                    if k not in ['qmtest.end_time','qmtest.start_time','qmtest.target']:
                        annotations[k] = v
                self.detailContext.outcomes.append(('<br>'.join(names),annotations))

            f = file(os.path.join(path,'detail_%03i.html' % i),'w')
            f.write(self.__tmplDetail.respond())
            f.close()
            i += 1

class TargetResult(object):
    """Encapsulates information about runs results for specific target.
    Attributes:
    name          Target name
    runs            List of RunResult instances ordered by seqno.
    """
    def __init__(self,target,runs):
        self.name = target
        self.runs = runs
        self.runs.sort(lambda runA,runB: cmp(runA.seqno,runB.seqno))

class RunResult(object):
    """Encapsulates information about run results.
    Attributes:
    annotations     Dictionary of run annotations (from qmtest context), key
                    is the annotation name.
    results         Dictionary of run results with test ID as key.
    target          Target name
    seqno           Run sequence number for this target
    """
    def __init__(self,result_stream):
        self.annotations = result_stream.GetAnnotations()
        self.results = {}
        for r in result_stream:
            r.owner = self
            self.results[r.GetId()] = r
        self.target = self.annotations.get("run_description",None)
        if self.target == None:
            self.target = Ask("Description for result file "+result_stream.file.name)
        self.seqno = self.annotations.get("run_seq",None)
        if self.seqno == None:
            self.seqno = Ask("Sequence number for result file "+result_stream.file.name)

class qaRepository(object):
    """Holds data about QMTest repository.
    
    Attributes:
    db -- Open QMTest Database instace (or None)
    test_map -- Dictionary that maps QA test IDs to qmtest test IDs (or None).
                Key is test ID, value is list of qmtest IDs for test variants.
    files_location -- Directory with suite special files (context variable).
    backup_location -- Directory with db backup files (context variable).
    suite_database_location -- Directory with spec. databases (context variable).
    result_archive -- Directory with test suite run results.
    """

    def __init__(self):
        self._del_db()

    def _set_db(self,path):
        if not qm_present:
            raise ImportError(QM_ERR)
        
        self._del_db()
        self.__db = qm.test.database.load_database(path)
        testids, suiteids = self.__db.GetSuite('').GetAllTestAndSuiteIds()
        # testmap is a dictionary that maps our QA test IDs to list of qmtest test IDs 
        # (i.e. test variants)
        self.test_map = {}
        for t in testids:
            self.test_map.setdefault(self.__db.GetItem('test',t).GetArguments()['test_id'],[]).append(t)
        
        # qmtest context variables
        self.files_location = os.path.join(self.__db.GetPath(),'files')
        self.backup_location = os.path.join(self.__db.GetPath(),'fbk')
        self.suite_database_location = os.path.join(self.__db.GetPath(),'fdb')
        self.result_archive = os.path.join(self.__db.GetPath(),'archive')
        if not os.path.exists(self.result_archive):
            os.mkdir(self.result_archive)
    def _get_db(self):
        return self.__db
    def _del_db(self):
        self.__db = None
        self.test_map = None
        self.files_location = None
        self.backup_location = None
        self.suite_database_location = None
        self.result_archive = None

    db = LateBindingProperty(_get_db,_set_db,_del_db,"Primary QMTest test database.")

class qaTarget(object):
    """Holds data about QA target (Firebird version)
    
    Attributes:
    server_location -- Host for DSN (empty means local protocol, context variable).
    user_name -- FB user name (context variable).
    user_password -- FB user password (context variable).
    fbdir -- Firebird's home directory (or None).
    version -- Firebird version (obtained from server).
    arch -- Firebird architecture (SS, CS or SC).
    platform -- Firebird platform (Windows, Linux, MacOS, FreeBSD or UNKNOWN)
    cpuarch -- CPU architecture (32 or 64).
    isql_path -- Path to isql (context variable).
    gbak_path -- Path to gbak (context variable).
    gsec_path -- Path to gsec (context variable).
    gstat_path -- Path to gstat (context variable).
    gfix_path -- Path to gfix (context variable).
    gpre_path -- Path to gpre (context variable).
    security_db -- Path to security database (context variable).
    """

    def __init__(self):
        self.server_location = ''
        self.user_name = 'SYSDBA'
        self.user_password = 'masterkey'
        self.fbdir = None
        self.version = UNKNOWN
        self.arch = UNKNOWN
        if sys.platform == 'win32':
            self.platform = 'Windows'
        elif sys.platform.startswith('linux'):
            self.platform = 'Linux'
        elif sys.platform.startswith('darwin'):
            self.platform = 'MacOS'
        else:
            self.platform = UNKNOWN
        self.cpuarch = platform.architecture()[0][:2]
        # qmtest context variables
        self.isql_path = None
        self.gbak_path = None
        self.gsec_path = None
        self.gstat_path = None
        self.gfix_path = None
        self.gpre_path = None
        self.security_db = None

class qaRun(object):
    """Holds data about QA test suite run.
    
    Attributes:
    results -- Disctionary of results.
               Key is run_description+run_sequence, value is list of QMTest Results.
    run_description -- Short description to be used for test run.
                       Format '<Platform><cpuarch>-<arch>-<run_seq>'
    run_seq -- Sequential number for test run.
    context -- QMTest context to be used for test run.
    """

    def __init__(self):
        self.results = {}
        self.tests = []
        self.run_description = UNKNOWN
        self.run_seq = 0
        self.tempdir = os.path.join(tempfile.gettempdir(),'qa/')
        self.context = {'temp_directory': self.tempdir,
                        'database_location': self.tempdir,
                        'DATABASE_PATH': self.tempdir}

class qaPerson(object):
    """Holds data about QA person that performs QA session.
    
    Attributes:
    name -- User name (or UNKNOWN).
    id -- User ID (exactly two letters).
    """

    def __init__(self):
        self.id = 'XX'
        self.name = UNKNOWN
    
class qaController(object):
    """Helper class to exctend controller with QA specific data.
    
    Holds path to repository, test data, platform and target.
    """

    platforms = ['Windows','Linux','Solaris','FreeBSD',
                    'MacOS','HP-UX','Sinix-Z']
    outcomes = ['PASS','FAIL','UNTESTED','ERROR']
    
    def __init__(self):
        self.locals = struct()
        
        self.repository = qaRepository()
        self.locals.repository = self.repository
        self.target = qaTarget()
        self.locals.target = self.target
        self.run_data = qaRun()
        self.locals.run_data = self.run_data
        self.person = qaPerson()
        self.locals.person = self.person

    def _get_architecture(self):
        return self.__architecture
    def _set_architecture(self,arch):
        if arch in ['SS','CS','SC']:
            self.__architecture = arch
        else:
            raise ValueError("Firebird architecture must be 'SS', 'CS' or 'SC'")
    def _get_test_map(self):
        return self.__test_map

    def _get_run_description(self):
        return '%s%s-%s' % (self.platform,self.cpuarch,self.architecture)
    
def extendController(controller):
    """Controller (console) extension function (see setup.py).
    
    Create and add qaController instance to the controller and console namespace.
    """
    qac = qaController()
        
    controller._qac = qac
    controller.locals['qa'] = qac.locals

class helpFbQA(HelpProvider):
    """Help Provider for Firebird QA PowerConsole commands and other QA topics"""

    help_firebird_qa = """Supported QA commands:
    
    QA ENV
       Print current QA configuration parameters
    QA REPOSITORY path
       Set main QMTest repository to <path>
    QA PERSON name id
       Set QA person identification to <name> and <id>. Id must have two letters.
    QA TARGET {SS | SC | CS} {LOCAL | host[:port]} [PASSWORD password]
       Set Firebird version for testing. Firebird must run, so additional information
       could be obtained via service API. Optinal PASSWORD is password for SYSDBA.
    QA RUN [FORMAT {FULL | BRIEF | BATCH | STATS | NONE}] [ARCHIVE | test-name]
       Run test(s) from active repository against active target for active QA person.
       If ARCHIVE option is present, stores results file in repository archive,
       otherwise in current working directory.
    QA REPORT [<version>]
       Generates HTML report from test results files stored in repository archive.
       If <version> is omitted, it generates report for active target.
    QA LOAD | SAVE CONFIG [filename]
       Loads or saves the QA environment configuration to/from file. If filename
       is omitted, file ~/.pwc/qa.cfg is used. Not all environment values are stored!
       See documentation for details.
    
    Global namespace also contains 'qa' object for direct access to QA data:
    qa.repository -- Information about QMTest repository
    qa.person -- Information about QA person
    qa.target -- Information about tested Firebird version
    qa.run_data -- Information associated with test suite run(s)
    
    Type 'help <object>' for more information.
    """

class cmdQAPerson(Command):
    """This command sets QA PERSON data.
    """
    
    def __init__(self,controller):
        super(cmdQAPerson,self).__init__(controller)
        self.data = controller._qac.person

        # Grammar
        self.keyCmdName  = self._makeNameNode(Empty())
        person_name = (Word(alphas,alphanums+'_').setName('person name (one word or single quoted)') | sglQuotedString)
        person_id = Word(alphas,exact=2).setName('person ID (exactly two letters)')
        # QA PERSON name id
        self.keyQAPerson = self._makeNameNode(Combine(
            CaselessKeyword('QA') + White() +
            CaselessKeyword('PERSON')))
        self.cmdQAPerson = self.keyCmdName +  \
            self.keyQAPerson.suppress() + \
            person_name.setResultsName('name') + \
            person_id.setResultsName('id')
        self.cmdQAPerson.setParseAction(self._compile)

    def _getGrammar(self):
        return self.cmdQAPerson
    def _getCmdKeyword(self):
        return self.keyQAPerson
    def execute(self,name=None,id=None):
        self.data.name = name
        self.data.id = id

class cmdQATarget(Command):
    """This command sets QA TARGET data.
    """
    
    def __init__(self,controller):
        super(cmdQATarget,self).__init__(controller)
        self.data = controller._qac.target
        self.qac = controller._qac

        # Grammar
        self.keyCmdName  = self._makeNameNode(Empty())
        version = Word(nums,nums+'.')
        arch = oneOf('SS CS SC',caseless=True).setParseAction(upcaseTokens)
        platform = oneOf(' '.join(controller._qac.platforms))
        cpuarch = oneOf('32 64 PPC')

        # QA TARGET arch LOCAL | host[:port] [PASSWORD password]
        optHost = Combine(Word(alphanums,alphanums+'.').setName('host')+Optional(':'+self.INTEGER).setName('port'))
        optPassword = CaselessKeyword('password') + QuotedString("'").setResultsName('password')
        
        self.keyQATarget = self._makeNameNode(Combine(
            CaselessKeyword('QA') + White() +
            CaselessKeyword('TARGET')))
        self.cmdQATarget = self.keyCmdName +  \
            self.keyQATarget.suppress() + arch.setResultsName('arch') + \
            (CaselessKeyword('LOCAL') ^ (optHost)).setResultsName('host') + Optional(optPassword)
        ## QA TARGET arch version [platform] [cpuarch] path
        self.cmdQATarget.setParseAction(self._compile)

    def _getGrammar(self):
        return self.cmdQATarget
    def _getCmdKeyword(self):
        return self.keyQATarget
    def execute(self,arch,host,password='masterkey'):
        self.data.arch = arch
        try:
            if host != 'LOCAL':
                self.data.server_location = host+':'
                svc = fbservice.connect(host=host,password=password)
            else:
                self.data.server_location = ''
                svc = fbservice.connect(password=password)
            version = svc.getServerVersion()
            arch = svc.getArchitecture().lower()
            self.data.version = version[4:version.index(' ')]
            fbdir = svc.getHomeDir()
            self.data.fbdir = fbdir
            ext = ''
            if arch.find('windows') != -1:
                self.data.platform = 'Windows'
                ext = '.exe'
            elif arch.find('linux') != -1:
                self.data.platform = 'Linux'
            elif arch.find('darwin') != -1:
                self.data.platform = 'MacOS'
            self.data.user_password = password
            
            # qmtest context variables
            self.data.isql_path = os.path.join(fbdir,'bin','isql'+ext)
            self.data.gbak_path = os.path.join(fbdir,'bin','gbak'+ext)
            self.data.gsec_path = os.path.join(fbdir,'bin','gsec'+ext)
            self.data.gstat_path = os.path.join(fbdir,'bin','gstat'+ext)
            self.data.gfix_path = os.path.join(fbdir,'bin','gfix'+ext)
            self.data.gpre_path = os.path.join(fbdir,'bin','gpre'+ext)
            self.data.security_db = svc.getSecurityDatabasePath()

            self.qac.run_data.run_description = '%s%s %s' % (self.data.platform,
                self.data.cpuarch,self.data.arch)
            cntx = dict()
            cntx['server_location'] = self.data.server_location
            cntx['user_name'] = self.data.user_name
            cntx['user_password'] = self.data.user_password
            cntx['isql_path'] = self.data.isql_path
            cntx['gbak_path'] = self.data.gbak_path
            cntx['gsec_path'] = self.data.gsec_path
            cntx['gstat_path'] = self.data.gstat_path
            cntx['gfix_path'] = self.data.gfix_path
            cntx['gpre_path'] = self.data.gpre_path
            cntx['isc4_path'] = self.data.security_db
            cntx['run_description'] = self.qac.run_data.run_description
            self.qac.run_data.context.update(cntx)
            
        except Exception,e:
            self.controller.writeErr('%s:%s' % (e.__class__, e))

class cmdQARepository(Command):
    """description of command QA Repository"""

    def __init__(self,controller):
        super(cmdQARepository,self).__init__(controller)
        self.data = controller._qac.repository
        self.qac = controller._qac
        
        # Grammar
        self.keyCmdName  = self._makeNameNode(Empty())
        # QA REPOSITORY repository
        self.keyQARepository = self._makeNameNode(Combine(
            CaselessKeyword('QA') + White() +
            CaselessKeyword('REPOSITORY')))
        self.cmdQARepository = self.keyCmdName +  \
            self.keyQARepository.suppress() + self.FILENAME.setResultsName('repository') + \
            Optional(lineEnd.suppress())
        self.cmdQARepository.setParseAction(self._compile)

    def _getGrammar(self):
        return self.cmdQARepository
    def _getCmdKeyword(self):
        return self.keyQARepository
    def execute(self,repository):

        if not qm_present:
            self.writeErr("QMTest not installed.\nYou can download QMTest from www.qmtest.com.\n")
            return
        if qm.test.database.is_database(repository):
            self.data.db = repository
            self.write('%i tests in database %s' % (len(self.data.test_map.keys()),repository))

            cntx = dict()
            cntx['suite_database_location'] = self.data.suite_database_location+'/'
            cntx['backup_location'] = self.data.backup_location+'/'
            cntx['files_location'] = self.data.files_location+'/'
            self.qac.run_data.context.update(cntx)
        else:
            self.writeErr("Path is not a valid QMTest test database")


class cmdQAEnvironment(Command):
    """This command displays current configuration of QA variables."""

    def __init__(self,controller):
        super(cmdQAEnvironment,self).__init__(controller)
        self.qac = controller._qac

        # Grammar
        self.keyCmdName  = self._makeNameNode(Empty())
        # QA ENV
        self.keyQAEnvironment = self._makeNameNode(Combine(
            CaselessKeyword('QA') + White() +
            CaselessKeyword('ENV')))
        self.cmdQAEnvironment = self.keyCmdName +  \
            self.keyQAEnvironment
        self.cmdQAEnvironment.setParseAction(self._compile)

    def _getGrammar(self):
        return self.cmdQAEnvironment
    def _getCmdKeyword(self):
        return self.keyQAEnvironment
    def execute(self):
        report = []
        if qm_present:
            qmversion = qm.config.version
        else:
            qmversion = 'Not installed'

        repository = self.qac.repository
        person = self.qac.person
        target = self.qac.target
        run_data = self.qac.run_data

        if repository.db:
            unique_tests = len(repository.test_map.keys())
            total_tests = reduce(lambda x,y: x+len(y),repository.test_map.values(),0)
        else:
            unique_tests = 0
            total_tests = 0
        
        report.append('QMTest: %s\n' % qmversion)
        if self.qac.repository.db:
            report.append('QMTest test repository: %s\n' % repository.db.GetPath())
            report.append(' QA files in:           %s\n' % asString(repository.files_location))
            report.append(' QA db backup files in: %s\n' % asString(repository.backup_location))
            report.append(' QA databases in:       %s\n' % asString(repository.suite_database_location))
            report.append(' QA archive in:         %s\n' % asString(repository.result_archive))
            report.append('\n %i tests in %i files\n\n' % (unique_tests,total_tests))
        else:
            report.append('QMTest test repository: NOT DEFIEND\n')

        report.append('QA person: ')
        if self.qac.person.name != UNKNOWN:
            report.append('%s (%s)\n' % (person.name, person.id))
        else:
            report.append('NOT DEFINED\n')
        
        if target.version != UNKNOWN:
            report.append('Target:     %s%s %s %s\n' % (target.platform,target.cpuarch,target.arch,target.version))
            report.append(' Host:      %s\n' % target.server_location)
            report.append(' Location:  %s\n' % target.fbdir)
            report.append(' Connect:   %s/%s\n' % (target.user_name,target.user_password))
            report.append(' isql:      %s\n' % target.isql_path)
            report.append(' gbak:      %s\n' % target.gbak_path)
            report.append(' gsec:      %s\n' % target.gsec_path)
            report.append(' gstat:     %s\n' % target.gstat_path)
            report.append(' gfix:      %s\n' % target.gfix_path)
            report.append(' gpre:      %s\n' % target.gpre_path)
            report.append(' SecurityDB:%s\n' % target.security_db)
        else:
            report.append('Target: NOT DEFINED\n')
        report.append('\n')
        
        report.append('QMTest run context:\n')
        for item in run_data.context.items():
            report.append(' %s = %s\n' % item)
        
        map(self.write, report)


class cmdQARun(Command):
    """description of command QA RUN"""

    def __init__(self,controller):
        super(cmdQARun,self).__init__(controller)
        self.qac = controller._qac

        # Grammar
        self.keyCmdName  = self._makeNameNode(Empty())
        testName = Word(alphas,alphanums+'_-.')
        # QA RUN [FORMAT {FULL | BRIEF | BATCH | STATS | NONE}] [ARCHIVE | test-name]
        self.keyQARun = self._makeNameNode(Combine(
            CaselessKeyword('QA') + White() +
            CaselessKeyword('RUN')))
        self.cmdQARun = self.keyCmdName +  self.keyQARun + \
            Optional(CaselessKeyword('FORMAT') + 
                oneOf('FULL BRIEF BATCH STATS NONE',caseless=True).setResultsName('format')) + \
            Optional(CaselessKeyword('ARCHIVE').setResultsName('archive') | testName.setResultsName('test')) + \
            Optional(lineEnd.suppress())
        self.cmdQARun.setParseAction(self._compile)

    def _getGrammar(self):
        return self.cmdQARun
    def _getCmdKeyword(self):
        return self.keyQARun

    def _print_error_message(self,message):
        """Output an error message.
    
        'message' -- Structured text for the error message to emit.  The
        messing is emitted to the standard error stream with an
        identifying prefix."""
        
        prefix = "qmtest: error: "
        message = qm.structured_text.to_text(str(message),
                                             indent=len(prefix))
        message = prefix + message[len(prefix):]
        sys.stderr.write(message)
    def _compare_versions(self,lver, rver):
        """ Compare two versions, returns:
        0 for lver = rver
        -1 for lver < rver
        1 for lver > rver"""
        if lver == None:
            lver = "0"
        if rver == None:
            rver = "0"
        lverlist = lver.split(".")[:3]
        rverlist = rver.split(".")[:3]
        if len(rverlist) < len(lverlist):
            for i in range(len(lverlist)-len(rverlist)):
                rverlist.append("0")
        if len(rverlist) > len(lverlist):
            for i in range(len(rverlist)-len(lverlist)):
                lverlist.append("0")
        for i in range(len(rverlist)):
            if (int(rverlist[i]) > int(lverlist[i])):
                return -1
            if (int(rverlist[i]) < int(lverlist[i])):
                return 1
        return 0
    def execute(self,format='brief',archive=None,test=None):

        target = self.qac.target
        repository = self.qac.repository
        person = self.qac.person
        run_data = self.qac.run_data
        run_data.run_seq = 0
        
        # checks
        if self.qac.person.name == UNKNOWN:
            self.writeErr("QA Person not defined.\n")
            return
        if not self.qac.repository.db:
            self.writeErr("QA Repository not defined.\n")
            return
        if target.version == UNKNOWN:
            self.writeErr("QA Target not defined.\n")
            return
        
        if archive:
            run_data.run_seq += 1
            output_path = os.path.join(repository.result_archive,target.version)
        else:
            output_path = os.getcwd()
        if not os.path.isdir(output_path):
            os.makedirs(output_path)
        while True:
            output_file = os.path.join(output_path,
                '%s%s-%s-%s%i.qmr' % (target.platform,target.cpuarch,
                                       target.arch,person.id,run_data.run_seq))
            if archive and os.path.isfile(output_file):
                run_data.run_seq += 1
            else:
                break
        
        params = ['-D',self.qac.repository.db.GetPath(),'run','-f',format.lower(),
                  '-o',output_file,
                  '-c','run_seq=%s%i' % (person.id,run_data.run_seq)]
        # context variables
        for context_item in (('-c',('%s=%s' % (key,value))) for key,value in run_data.context.items()):
            params.extend(context_item)
        test_list = []
        # test names
        if test == None:
            for x,versions in repository.test_map.items():
                candidate = '0'
                test = None
                tests = (repository.db.GetItem('test',test_id) for test_id in versions)
                tests_for_platform = (t for t in tests 
                                      if target.platform in t.GetArguments()['target_platform'] 
                                      or t.GetArguments()['target_platform'] =='All')
                
                for t in tests_for_platform:
                    version = t.GetArguments()['target_version']
                    if self._compare_versions(target.version,version) >= 0:
                        if self._compare_versions(candidate,version) < 0:
                            candidate = version
                            test = t
                if test:
                    test_list.append(test.GetId())
        else:
            test_list.append(test)
        params.extend(test_list)
        
        try:
            command = QMTest(params,sys.argv[0])
            try:
                pass
                exit_code = command.Execute()
            except qm.cmdline.CommandError, msg:
                self._print_error_message(msg)
            except qm.common.QMException, msg:
                self._print_error_message(msg)
            except NotImplementedError:
                exc_info = sys.exc_info()
                method_name = traceback.extract_tb(exc_info[2])[-1][2]
                self._print_error_message(qm.message("not implemented",
                                               method_name = method_name))
                self.writeErr(qm.common.format_traceback(exc_info))
            except KeyboardInterrupt:
                self.writeErr("\nqmtest: Interrupted.\n")
                raise
            except qm.platform.SignalException, se:
                # SIGTERM indicates a request to shut down.  Other signals
                # should be handled earlier.
                if se.GetSignalNumber() != signal.SIGTERM:
                    raise
        finally:
            del qm.test.cmdline._the_qmtest
            gc.collect()

class cmdQAReport(Command):
    """description of command QA REPORT"""

    def __init__(self,controller):
        super(cmdQAReport,self).__init__(controller)
        self.qac = controller._qac

        # Grammar
        self.keyCmdName  = self._makeNameNode(Empty())
        # QA REPORT 
        self.keyQAReport = self._makeNameNode(Combine(
            CaselessKeyword('QA') + White() +
            CaselessKeyword('REPORT')) + Optional(self.ARG.setResultsName('version')))
        self.cmdQAReport = self.keyCmdName +  self.keyQAReport.suppress()
        self.cmdQAReport.setParseAction(self._compile)

    def _getGrammar(self):
        return self.cmdQAReport
    def _getCmdKeyword(self):
        return self.keyQAReport

    def execute(self,version=None):
        target = self.qac.target
        repository = self.qac.repository

        if not self.qac.repository.db:
            self.writeErr("QA Repository not defined.")
            return
        if version == None:
            if target.version == UNKNOWN:
                self.writeErr("QA Target not defined.")
                return
            else:
                version = target.version

        version_dir = os.path.join(repository.result_archive,version)
        report_dir = os.path.join(version_dir,'report')
        if not os.path.isdir(report_dir):
            os.makedirs(report_dir)
        # create list of .qmr files
        qmrfiles = []
        for filename in glob.glob(os.path.join(version_dir,'*.qmr')):
            print 'Input:',filename
            qmrfiles.append(filename)
        
        # get list of runs
        runs = []
        for result_file in qmrfiles:
            try:
                result_stream = base.load_results(open(result_file, "rb"), None)
                runs.append(RunResult(result_stream))
            except IOError, e:
                raise PythonException("Error reading '%s'"%result_file, IOError, e)
            except xml.sax.SAXException, e:
                raise PythonException("Error loading '%s'"%result_file, xml.sax.SAXException, e)
        # get list of targets
        targets = []
        for run in runs:
            if run.target not in targets:
                targets.append(run.target)
        # create list of target results
        targetruns = []
        for target in targets:
            targetruns.append(TargetResult(target,[run for run in runs if run.target == target]))
        
        report = Report(targetruns,version)
    
        report.generateHTMLReport(report_dir)

class cmdQAConfig(Command):
    """description of command QA Config"""

    def __init__(self,controller):
        super(cmdQAConfig,self).__init__(controller)
        self.qac = controller._qac

        # Grammar
        self.keyCmdName  = self._makeNameNode(Empty())
        # QA LOAD | SAVE CONFIG [<filename>]
        self.keyQAConfig = self._makeNameNode(CaselessKeyword('QA') + White() + \
            (CaselessKeyword('LOAD') | CaselessKeyword('SAVE')))
        self.cmdQAConfig = self.keyCmdName +  \
            CaselessKeyword('QA') + White() + \
            (CaselessKeyword('LOAD') | CaselessKeyword('SAVE')).setResultsName('action') + \
            CaselessKeyword('CONFIG') + Optional(self.FILENAME.setResultsName('cfgfile')) + \
            Optional(lineEnd.suppress())
        self.cmdQAConfig.setParseAction(self._compile)

    def _getGrammar(self):
        return self.cmdQAConfig
    def _getCmdKeyword(self):
        return self.keyQAConfig

    def execute(self,action,cfgfile=None):
        
        repository = self.qac.repository
        person = self.qac.person
        target = self.qac.target
        run_data = self.qac.run_data

        def handleItem(section,key):
            if key == 'repository':
                repository.db = section[key]
            elif key == 'platform':
                target.platform = section[key]
            elif key == 'cpuarch':
                target.cpuarch = section[key]
            elif key == 'files_location':
                repository.files_location = section[key]
            elif key == 'backup_location':
                repository.backup_location = section[key]
            elif key == 'suite_database_location':
                repository.suite_database_location = section[key]
            elif key == 'result_archive':
                repository.result_archive = section[key]
            elif key == 'person_name':
                person.name = section[key]
            elif key == 'person_id':
                person.id = section[key]
            elif key == 'temp_directory':
                run_data.context[key] = section[key]
            elif key == 'DATABASE_PATH':
                run_data.context[key] = section[key]
            elif key == 'database_location':
                run_data.context[key] = section[key]

        if not cfgfile:
            cfgfile = os.path.expanduser('~/.pwc/qa.cfg')
        if action == 'LOAD':
            if os.path.isfile(cfgfile):
                cfg = ConfigObj(cfgfile)
                cfg.walk(handleItem)
            else:
                self.writeErr("Config file '%s' not found." % cfgfile)
        elif action == 'SAVE':
            cfg = ConfigObj()
            cfg.filename = cfgfile
            (d,f) = os.path.split(cfgfile)
            if not os.path.isdir(d):
                os.makedirs(d)
            cfg.initial_comment = ['PowerConsole QA Pack configuration','']
            if repository.db:
                cfg['repository'] = repository.db.GetPath()
                cfg.inline_comments['repository'] = 'Path to QMTest test database'

            if repository.files_location:
                cfg['files_location'] = repository.files_location
                cfg.inline_comments['files_location'] = 'Path to suite supplementary files'
            if repository.files_location:
                cfg['backup_location'] = repository.files_location
                cfg.inline_comments['backup_location'] = 'Path to suite database backup files'
            if repository.suite_database_location:
                cfg['suite_database_location'] = repository.suite_database_location
                cfg.inline_comments['suite_database_location'] = 'Path to suite database files'
            if repository.result_archive:
                cfg['result_archive'] = repository.result_archive
                cfg.inline_comments['result_archive'] = 'Path to QA run results archive'
            
            if person.name != UNKNOWN:
                cfg['person_name'] = person.name
                cfg.inline_comments['person_name'] = 'QA person name'
                cfg['person_id'] = person.id
                cfg.inline_comments['person_id'] = 'QA person ID'

            cfg['platform'] = target.platform
            cfg.inline_comments['platform'] = 'Target platform'
            cfg['cpuarch'] = target.cpuarch
            cfg.inline_comments['cpuarch'] = 'Target CPU architecture'

            cfg['temp_directory'] = run_data.context['temp_directory']
            cfg.inline_comments['temp_directory'] = 'Path to directory for temporary files'
            cfg['DATABASE_PATH'] = run_data.context['DATABASE_PATH']
            cfg.inline_comments['DATABASE_PATH'] = 'Path to directory where databases are created'
            cfg['database_location'] = run_data.context['database_location']
            cfg.inline_comments['database_location'] = 'Path to directory where databases are created'
            cfg.write()

class packageFirebirdQA(ExtensionPackage):
    """Firebird QA Package handler object for PowerConsole.
    """

    def __init__(self):
        super(packageFirebirdQA,self).__init__('pwcfbqa')
    
    def controller_extensions(self):
        return [extendController]
    def commands(self):
        return [cmdQAPerson,cmdQATarget,cmdQARepository,cmdQAEnvironment,
                cmdQARun,cmdQAReport,cmdQAConfig]
    def help_providers(self):
        return [helpFbQA]
